from amulet_map_editor import lang
from amulet_map_editor.api.opengl import check_opengl

check_opengl()

from amulet_map_editor.programs.edit.edit import EditExtension

export = {"name": lang.get("program_3d_edit.tab_name"), "ui": EditExtension}



#!/usr/bin/env python
#
# Copyright (c) 2014 Matthew Borgerson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Texture Atlas and Map File Generation Utility Classes"""

from PIL import Image
import math
from typing import Dict, Tuple, List, Optional, Generator
from amulet_map_editor import log

DESCRIPTION = """Packs many smaller images into one larger image, a Texture
Atlas. A companion file (.map), is created that defines where each texture is
mapped in the atlas."""


class AtlasTooSmall(Exception):
    pass


class Packable(object):
    """A two-dimensional object with position information."""

    def __init__(self, width: int, height: int):
        self._x = 0
        self._y = 0
        self._width = width
        self._height = height

    @property
    def x(self) -> int:
        return self._x

    @x.setter
    def x(self, value: int):
        self._x = value

    @property
    def y(self) -> int:
        return self._y

    @y.setter
    def y(self, value: int):
        self._y = value

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def perimeter(self) -> int:
        return 2 * self._width + 2 * self._height


class PackRegion(object):
    """A region that two-dimensional Packable objects can be packed into."""

    def __init__(self, x: int, y: int, width: int, height: int):
        """Constructor."""
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._sub1: Optional[PackRegion] = None
        self._sub2: Optional[PackRegion] = None
        self._packable: Optional[Packable] = None

    @property
    def x(self) -> int:
        return self._x

    @property
    def y(self) -> int:
        return self._y

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def packable(self) -> Packable:
        return self._packable

    def get_all_packables(self):
        """Returns a list of all Packables in this region and sub-regions."""
        if self._packable:
            return (
                [self._packable]
                + self._sub1.get_all_packables()
                + self._sub2.get_all_packables()
            )
        return []

    def pack(self, packable: Packable, border: int):
        """Pack 2D packable into this region."""
        if not self._packable:
            # Is there room to pack this?
            if (packable.width + border * 2 > self._width) or (
                packable.height + border * 2 > self._height
            ):
                return False

            # Pack
            self._packable = packable

            # Set x, y on Packable
            self._packable.x = self._x + border
            self._packable.y = self._y + border

            # Create sub-regions
            self._sub1 = PackRegion(
                self._x,
                self._y + self._packable.height + border * 2,
                self._packable.width + border * 2,
                self._height - self._packable.height - border * 2,
            )
            self._sub2 = PackRegion(
                self._x + self._packable.width + border * 2,
                self._y,
                self._width - self._packable.width - border * 2,
                self._height,
            )
            return True

        # Pack into sub-region
        return self._sub1.pack(packable, border) or self._sub2.pack(packable, border)


class Frame(Packable):
    """An image file that can be packed into a PackRegion."""

    def __init__(self, filename: str):
        self._filename = filename

        # Determine frame dimensions
        image: Image.Image = Image.open(filename)
        self._image: Image.Image = image.copy()
        image.close()

        width, height = self._image.size

        super(Frame, self).__init__(width, height)

    @property
    def filename(self) -> str:
        return self._filename

    def draw(self, image: Image.Image, border: int):
        """Draw this frame into another Image."""
        if border:
            image.paste(
                self._image.resize(tuple(s + border * 2 for s in self._image.size)),
                (self.x - border, self.y - border),
            )
        image.paste(self._image, (self.x, self.y))


class Texture(object):
    """A collection of one or more frames."""

    def __init__(self, name: str, frames: List[Frame]):
        self._name = name
        self._frames: List[Frame] = frames

    @property
    def name(self) -> str:
        return self._name

    @property
    def frames(self) -> List[Frame]:
        return self._frames


class TextureAtlas(PackRegion):
    """Texture Atlas generator."""

    def __init__(self, width: int, height: int, border: int = 0):
        super(TextureAtlas, self).__init__(0, 0, width, height)
        self._textures: List[Texture] = []
        self._border = border

    @property
    def textures(self) -> List[Texture]:
        return self._textures

    def pack(self, texture: Texture):
        """Pack a Texture into this atlas."""
        self._textures.append(texture)
        for frame in texture.frames:
            if not super(TextureAtlas, self).pack(frame, self._border):
                raise AtlasTooSmall("Failed to pack frame %s" % frame.filename)

    def to_dict(self) -> Dict[str, Tuple[int, int, int, int]]:
        return {
            tex.name: (
                tex.frames[0].x / self.width,
                tex.frames[0].y / self.height,
                (tex.frames[0].x + tex.frames[0].width) / self.width,
                (tex.frames[0].y + min(tex.frames[0].height, tex.frames[0].width))
                / self.height,
            )
            for tex in self.textures
        }

    def generate(self, mode: str) -> Image.Image:
        """Generates the final texture atlas."""
        out = Image.new(mode, (self.width, self.height))
        for t in self._textures:
            for f in t.frames:
                f.draw(out, self._border)
        return out

    def write(self, filename: str, mode: str):
        """Generates and saves the final texture atlas."""
        out = self.generate(mode)
        out.save(filename)


class TextureAtlasMap(object):
    """Texture Atlas Map file generator."""

    def __init__(self, atlas):
        self._atlas = atlas

    def write(self, fd):
        """Writes the texture atlas map file into file object fd."""
        raise Exception("Not Implemented")


def create_atlas(
    texture_tuple: Tuple[str, ...]
) -> Tuple[Image.Image, Dict[str, Tuple[float, float, float, float]]]:
    atlas_iter = create_atlas_iter(texture_tuple)
    try:
        while True:
            next(atlas_iter)
    except StopIteration as e:
        return e.value


def create_atlas_iter(
    texture_tuple: Tuple[str, ...]
) -> Generator[
    float,
    None,
    Tuple[Image.Image, Dict[str, Tuple[float, float, float, float]]],
]:
    log.info("Creating texture atlas")
    # Parse texture names
    textures = []
    for texture_index, texture in enumerate(texture_tuple):
        if not texture_index % 100:
            yield texture_index / (len(texture_tuple) * 2)
        # Look for a texture name
        name, frames = texture, [texture]

        # Build frame objects
        frames = [Frame(f) for f in frames]

        # Add frames to texture object list
        textures.append(Texture(name, frames))

    # Sort textures by perimeter size in non-increasing order
    textures = sorted(textures, key=lambda i: i.frames[0].perimeter, reverse=True)

    height = 0
    width = 0
    pixels = 0
    for t in textures:
        for f in t.frames:
            height = max(f.height, height)
            width = max(f.width, width)
            pixels += f.height * f.width

    size = max(height, width, 1 << (math.ceil(pixels**0.5) - 1).bit_length())

    atlas_created = False
    atlas = None
    while not atlas_created:
        try:
            # Create the atlas and pack textures in
            log.info(f"Trying to pack textures into image of size {size}x{size}")
            atlas = TextureAtlas(size, size)

            for texture_index, texture in enumerate(textures):
                if not texture_index % 30:
                    yield 0.5 + texture_index / (len(textures) / 2)
                atlas.pack(texture)
            atlas_created = True
        except AtlasTooSmall:
            log.info(f"Image was too small. Trying with a larger area")
            size *= 2

    log.info(f"Successfully packed textures into an image of size {size}x{size}")

    texture_atlas = atlas.generate("RGBA")

    texture_bounds = atlas.to_dict()
    texture_bounds = {
        texture_path: texture_bounds[texture_path] for texture_path in texture_tuple
    }

    log.info("Finished creating texture atlas")
    return texture_atlas, texture_bounds



from typing import TYPE_CHECKING
import wx
import os

from amulet.api.selection import SelectionGroup
from amulet.api.errors import ChunkLoadError
from amulet.api.data_types import Dimension, OperationReturnType
from amulet.level.formats.construction import ConstructionFormatWrapper

from amulet_map_editor.api.wx.ui.version_select import VersionSelect
from amulet_map_editor.programs.edit.api.operations import (
    SimpleOperationPanel,
    OperationError,
)

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class ExportConstruction(SimpleOperationPanel):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleOperationPanel.__init__(self, parent, canvas, world, options_path)

        options = self._load_options({})

        self._file_picker = wx.FilePickerCtrl(
            self,
            path=options.get("path", ""),
            wildcard="Construction file (*.construction)|*.construction",
            style=wx.FLP_USE_TEXTCTRL | wx.FLP_SAVE | wx.FLP_OVERWRITE_PROMPT,
        )
        self._sizer.Add(self._file_picker, 0, wx.ALL | wx.CENTER, 5)
        self._version_define = VersionSelect(
            self,
            world.translation_manager,
            options.get("platform", None) or world.level_wrapper.platform,
            allow_universal=False,
        )
        self._sizer.Add(self._version_define, 0, wx.CENTRE, 5)
        self._add_run_button("Export")
        self.Layout()

    def disable(self):
        self._save_options(
            {
                "path": self._file_picker.GetPath(),
                "platform": self._version_define.platform,
                "version": self._version_define.version_number,
            }
        )

    def _operation(
        self, world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
    ) -> OperationReturnType:
        path = self._file_picker.GetPath()
        platform = self._version_define.platform
        version = self._version_define.version_number
        if isinstance(path, str) and platform and version:
            wrapper = ConstructionFormatWrapper(path)
            if wrapper.exists:
                response = wx.MessageDialog(
                    self,
                    f"A file is already present at {path}. Do you want to continue?",
                    style=wx.YES | wx.NO,
                ).ShowModal()
                if response == wx.ID_CANCEL:
                    return
            wrapper.create_and_open(platform, version, selection, True)
            wrapper.translation_manager = world.translation_manager
            wrapper_dimension = wrapper.dimensions[0]
            chunk_count = len(list(selection.chunk_locations()))
            yield 0, f"Exporting {os.path.basename(path)}"
            for chunk_index, (cx, cz) in enumerate(selection.chunk_locations()):
                try:
                    chunk = world.get_chunk(cx, cz, dimension)
                    wrapper.commit_chunk(chunk, wrapper_dimension)
                except ChunkLoadError:
                    continue
                yield (chunk_index + 1) / chunk_count
            wrapper.save()
            wrapper.close()
        else:
            raise OperationError(
                "Please specify a save location and version in the options before running."
            )


export = {
    "name": "\tExport Construction",  # the name of the plugin
    "operation": ExportConstruction,  # the UI class to display
}



import wx
import wx.lib.scrolledpanel
from typing import Tuple, Optional, Dict

import PyMCTranslate
from amulet.api.block import PropertyType, Block
from amulet.api.block_entity import BlockEntity

from amulet_map_editor.api.wx.ui.base_define import BaseDefine
from amulet_map_editor.api.wx.ui.block_select import BlockSelect

from amulet_map_editor.api.wx.ui.block_select.properties import (
    PropertySelect,
    WildcardSNBTType,
    EVT_PROPERTIES_CHANGE,
)


class BlockDefine(BaseDefine):
    def __init__(
        self,
        parent,
        translation_manager: PyMCTranslate.TranslationManager,
        orientation=wx.VERTICAL,
        platform: str = None,
        version_number: Tuple[int, int, int] = None,
        force_blockstate: bool = None,
        namespace: str = None,
        block_name: str = None,
        properties: PropertyType = None,
        wildcard_properties=False,
        show_pick_block: bool = False,
        **kwargs,
    ):
        super().__init__(
            parent,
            translation_manager,
            BlockSelect,
            orientation,
            platform,
            version_number,
            namespace,
            default_name=block_name,
            show_pick=show_pick_block,
            force_blockstate=force_blockstate,
            **kwargs,
        )

        right_sizer = wx.BoxSizer(wx.VERTICAL)
        if orientation == wx.HORIZONTAL:
            self._sizer.Add(right_sizer, 1, wx.EXPAND | wx.LEFT, 5)
        else:
            self._sizer.Add(right_sizer, 1, wx.EXPAND | wx.TOP, 5)

        self._property_picker = PropertySelect(
            self,
            translation_manager,
            self._version_picker.platform,
            self._version_picker.version_number,
            self._version_picker.force_blockstate,
            self._picker.namespace,
            self._picker.name,
            properties,
            wildcard_properties,
        )
        right_sizer.Add(self._property_picker, 1, wx.EXPAND)
        self._property_picker.Bind(EVT_PROPERTIES_CHANGE, self._on_property_change)

        self.SetSizerAndFit(self._sizer)
        self.Layout()

    def _on_picker_change(self, evt):
        self._update_properties()
        evt.Skip()

    def _on_property_change(self, evt):
        self.Layout()
        evt.Skip()

    def _update_properties(self):
        self._property_picker.version_block = (
            self._version_picker.platform,
            self._version_picker.version_number,
            self._version_picker.force_blockstate,
            self._picker.namespace,
            self._picker.name,
        )

    @property
    def force_blockstate(self) -> bool:
        return self._version_picker.force_blockstate

    @force_blockstate.setter
    def force_blockstate(self, force_blockstate: bool):
        self._version_picker.force_blockstate = force_blockstate

    @property
    def block_name(self) -> str:
        return self._picker.name

    @block_name.setter
    def block_name(self, block_name: str):
        self._picker.name = block_name

    @property
    def str_properties(self) -> Dict[str, "WildcardSNBTType"]:
        return self._property_picker.str_properties

    @str_properties.setter
    def str_properties(self, str_properties: Dict[str, "WildcardSNBTType"]):
        self._property_picker.str_properties = str_properties

    @property
    def properties(self) -> PropertyType:
        return self._property_picker.properties

    @properties.setter
    def properties(self, properties: PropertyType):
        self._property_picker.properties = properties

    @property
    def block(self) -> Block:
        return Block(self.namespace, self.block_name, self.properties)

    @block.setter
    def block(self, block: Block):
        self._picker.set_namespace(block.namespace)
        self._picker.set_name(block.base_name)
        self._update_properties()
        self.properties = block.properties

    @property
    def block_entity(self) -> Optional[BlockEntity]:
        return None  # TODO

    @block_entity.setter
    def block_entity(self, block_entity: Optional[BlockEntity]):
        if block_entity is not None:
            pass  # TODO

    @property
    def universal_block(self) -> Tuple[Block, Optional[BlockEntity]]:
        return self._translation_manager.get_version(
            self.platform, self.version_number
        ).block.to_universal(self.block, self.block_entity, self.force_blockstate)[:2]

    @universal_block.setter
    def universal_block(self, universal_block: Tuple[Block, Optional[BlockEntity]]):
        block, block_entity = universal_block
        v_block, v_block_entity = self._translation_manager.get_version(
            self.platform, self.version_number
        ).block.from_universal(block, block_entity, self.force_blockstate)[:2]
        if isinstance(v_block, Block):
            self.block = v_block
            self.block_entity = v_block_entity


if __name__ == "__main__":

    def main():
        app = wx.App()
        translation_manager = PyMCTranslate.new_translation_manager()
        dialog = wx.Dialog(None, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        sizer = wx.BoxSizer()
        dialog.SetSizer(sizer)
        sizer.Add(
            BlockDefine(dialog, translation_manager, wx.HORIZONTAL),
            1,
            wx.ALL | wx.EXPAND,
            5,
        )
        dialog.Show()
        dialog.Fit()
        app.MainLoop()

    main()



from __future__ import annotations

from typing import Optional
import urllib.request
import threading
import json
import webbrowser

import wx
from packaging.version import Version

from amulet_map_editor import lang

URL = "http://api.github.com/repos/Amulet-Team/Amulet-Map-Editor/releases"

_EVT_UPDATE_CHECK = wx.NewEventType()
EVT_UPDATE_CHECK = wx.PyEventBinder(_EVT_UPDATE_CHECK, 1)


class UpdateEvent(wx.PyCommandEvent):
    def __init__(self, etype, eid, new_version: str):
        wx.PyCommandEvent.__init__(self, etype, eid)
        self._new_version = new_version

    def GetVersion(self) -> str:
        return self._new_version


class UpdateDialog(wx.Dialog):
    def __init__(self, parent, current_version: str, new_version: str):
        super().__init__(parent)

        sizer_1 = wx.BoxSizer(wx.VERTICAL)

        static_text_1 = wx.StaticText(
            self, label=lang.get("update_check.newer_version_released")
        )
        static_text_2 = wx.StaticText(
            self, label=f"{lang.get('update_check.new_version')} v{new_version}"
        )
        static_text_3 = wx.StaticText(
            self, label=f"{lang.get('update_check.current_version')} v{current_version}"
        )

        sizer_1.Add(static_text_1, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        sizer_1.Add(static_text_2, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        sizer_1.Add(static_text_3, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)

        update_button = wx.Button(self, label=lang.get("update_check.update"))
        ok_button = wx.Button(self, label=lang.get("update_check.ok"))

        sizer_2.Add(update_button, 0, wx.ALL, 5)
        sizer_2.Add((0, 0), 1, wx.EXPAND, 5)
        sizer_2.Add(ok_button, 0, wx.ALL, 5)

        sizer_1.Add(sizer_2, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(sizer_1)

        self.Centre()

        update_button.Bind(
            wx.EVT_BUTTON, lambda evt: self.goto_download_page(new_version, evt)
        )
        ok_button.Bind(wx.EVT_BUTTON, lambda evt: self.Close())

    @staticmethod
    def goto_download_page(new_version, _):
        webbrowser.open(
            f"https://github.com/Amulet-Team/Amulet-Map-Editor/releases/tag/{new_version}"
        )


def _is_compatible(current_version: Version, release_version: Version) -> bool:
    """The release version is only compatible with the current version if it is newer and more stable."""
    # release > beta > beta.dev > alpha.dev
    #                > alpha > alpha.dev
    def get_release_stage(version: Version) -> int:
        if version.pre is None:
            return 3
        else:
            return {"a": 0, "b": 1, "rc": 2}[version.pre[0]]

    return (
        # The release version is newer than the current version
        release_version > current_version
        # the pre-release stage is newer or the same
        and get_release_stage(release_version) >= get_release_stage(current_version)
        # dev builds should only be suggested if the current version is a dev build
        and (release_version.dev is None or current_version.dev is not None)
    )


def _get_newest_version(url: str, current_version_str: str) -> Optional[str]:
    """Find a newer but compatible release than this one if one exists."""

    try:
        current_version = Version(current_version_str)

        if current_version.local is not None:
            # if the version has a local extension (eg. "+0.gee5780.dirty")
            print("Running from source, not showing update dialog")
            return

        conn = urllib.request.urlopen(url, timeout=5)
        data = conn.read()
        data = json.loads(data)

        # iterate through all release versions starting with the newest
        for release_version, release_data in sorted(
            map(lambda d: (Version(d["tag_name"]), d), data),
            key=lambda t: t[0],
            reverse=True,
        ):
            if _is_compatible(current_version, release_version):
                return str(release_version)

    except Exception:
        pass


def _check_for_update(
    listening_parent: wx.EvtHandler, url: str, current_version_str: str
):
    """Check if there is a newer release and post an UpdateEvent if there is."""
    release_version = _get_newest_version(url, current_version_str)
    if release_version is not None:
        evt = UpdateEvent(_EVT_UPDATE_CHECK, -1, str(release_version))
        wx.PostEvent(listening_parent, evt)


def check_for_update(listening_parent: wx.EvtHandler, current_version_str: str):
    update_thread = threading.Thread(
        target=_check_for_update, args=(listening_parent, URL, current_version_str)
    )
    update_thread.start()



import wx
from typing import Callable, Dict, Any, TYPE_CHECKING, Sequence

from amulet_map_editor.api.wx.util.validators import IntValidator

from amulet.api.data_types import OperationReturnType
from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas
    from amulet.api.level import BaseLevel

FixedOperationType = Callable[
    ["BaseLevel", "Dimension", "SelectionGroup", Dict[str, Any]], OperationReturnType
]


class FixedFunctionUI(wx.Panel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
        operation: FixedOperationType,
        options: Dict[str, Any],
    ):
        wx.Panel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)
        self._operation = operation

        self.Hide()
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)
        self._options_sizer = wx.BoxSizer(wx.VERTICAL)
        self._sizer.Add(self._options_sizer)
        self._run_button = wx.Button(self, label="Run Operation")
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self._options: Dict[str, wx.Window] = {}
        self._create_options(options)

        self.Layout()
        self.Show()

    def _create_options(self, options: Dict[str, Sequence]):
        create_functions: Dict[str, Callable[[str, Sequence], None]] = {
            "label": self._create_label,
            "bool": self._create_bool,
            "int": self._create_int,
            "float": self._create_float,
            "str": self._create_string,
            "str_choice": self._create_str_choice,
            "file_open": self._create_file_open_picker,
            "file_save": self._create_file_save_picker,
            "directory": self._create_directory_picker,
        }
        for option_name, option in options.items():
            if not (isinstance(option, (list, tuple)) and option):
                continue
            option_type, option = option[0], option[1:]
            if option_type not in create_functions:
                continue
            create_functions[option_type](option_name, option)

    def _create_label(self, option_name: str, options: Sequence):
        label = wx.StaticText(self, label=option_name)
        self._options_sizer.Add(label, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

    def _create_horizontal_options_sizer(self, label) -> wx.BoxSizer:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        label = wx.StaticText(self, label=label)
        sizer.Add(label, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        self._options_sizer.Add(sizer, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)
        return sizer

    def _create_bool(self, option_name: str, options: Sequence):
        if options and not isinstance(options[0], bool):
            return
        sizer = self._create_horizontal_options_sizer(option_name)
        option = wx.CheckBox(self)
        sizer.Add(option)
        if options:
            option.SetValue(options[0])
        self._options[option_name] = option

    def _create_int(self, option_name: str, options: Sequence):
        if len(options) in {0, 1, 3} and all(isinstance(o, int) for o in options):
            sizer = self._create_horizontal_options_sizer(option_name)
            if len(options) == 0:
                option = wx.SpinCtrl(
                    self,
                    min=-30_000_000,
                    max=30_000_000,
                    initial=0,
                )
            elif len(options) == 1:
                option = wx.SpinCtrl(
                    self,
                    min=-30_000_000,
                    max=30_000_000,
                    initial=options[0],
                )
            elif len(options) == 3:
                option = wx.SpinCtrl(
                    self,
                    min=min(options[1:3]),
                    max=max(options[1:3]),
                    initial=options[0],
                )
            else:
                return  # should not get here
            option.SetValidator(IntValidator())
            sizer.Add(option)
            self._options[option_name] = option

    def _create_float(self, option_name: str, options: Sequence):
        if len(options) in {0, 1, 3} and all(
            isinstance(o, (int, float)) for o in options
        ):
            sizer = self._create_horizontal_options_sizer(option_name)
            if len(options) == 0:
                option = wx.SpinCtrlDouble(
                    self,
                    min=-30_000_000,
                    max=30_000_000,
                    initial=0,
                )
            elif len(options) == 1:
                option = wx.SpinCtrlDouble(
                    self,
                    min=-30_000_000,
                    max=30_000_000,
                    initial=options[0],
                )
            elif len(options) == 3:
                option = wx.SpinCtrlDouble(
                    self,
                    min=min(options[1:3]),
                    max=max(options[1:3]),
                    initial=options[0],
                )
            else:
                return  # should not get here
            sizer.Add(option)
            self._options[option_name] = option

    def _create_string(self, option_name: str, options: Sequence):
        if options and not isinstance(options[0], str):
            return
        sizer = self._create_horizontal_options_sizer(option_name)
        option = wx.TextCtrl(self)
        sizer.Add(option)
        if options:
            option.SetValue(options[0])
        self._options[option_name] = option

    def _create_str_choice(self, option_name: str, options: Sequence):
        if not (options and all(isinstance(o, str) for o in options)):
            return
        sizer = self._create_horizontal_options_sizer(option_name)
        option = wx.Choice(self, choices=options)
        option.SetSelection(0)
        sizer.Add(option)
        self._options[option_name] = option

    def _create_file_save_picker(self, option_name: str, options: Sequence):
        sizer = self._create_horizontal_options_sizer(option_name)
        option = wx.FilePickerCtrl(self, style=wx.FLP_SAVE | wx.FLP_USE_TEXTCTRL)
        sizer.Add(option)
        self._options[option_name] = option

    def _create_file_open_picker(self, option_name: str, options: Sequence):
        sizer = self._create_horizontal_options_sizer(option_name)
        option = wx.FilePickerCtrl(self, style=wx.FLP_OPEN | wx.FLP_USE_TEXTCTRL)
        sizer.Add(option)
        self._options[option_name] = option

    def _create_directory_picker(self, option_name: str, options: Sequence):
        sizer = self._create_horizontal_options_sizer(option_name)
        option = wx.DirPickerCtrl(self, style=wx.DIRP_USE_TEXTCTRL)
        sizer.Add(option)
        self._options[option_name] = option

    def _get_values(self) -> Dict[str, Any]:
        options = {}
        for key, window in self._options.items():
            if isinstance(
                window,
                (
                    wx.CheckBox,
                    wx.SpinCtrl,
                    wx.SpinCtrlDouble,
                    wx.TextCtrl,
                ),
            ):
                options[key] = window.GetValue()
            elif isinstance(window, wx.Choice):
                options[key] = window.GetString(window.GetSelection())
            elif isinstance(window, (wx.FilePickerCtrl, wx.DirPickerCtrl)):
                options[key] = window.GetPath()
        return options

    def _run_operation(self, evt):
        self.canvas.run_operation(
            lambda: self._operation(
                self.world,
                self.canvas.dimension,
                self.canvas.selection.selection_group,
                self._get_values(),
            )
        )



import threading

import wx
from typing import List, Callable, Tuple, Type, Union, Optional
import traceback
import importlib
import pkgutil

from amulet.api.errors import LoaderNoneMatched
from amulet import load_level

from amulet_map_editor import programs, log, lang
from amulet_map_editor.api.datatypes import MenuData
from amulet_map_editor.api.framework.pages import BasePageUI
from amulet_map_editor.api.framework.programs import BaseProgram, AboutProgram
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog

_extensions: List[Tuple[str, Type[BaseProgram]]] = []
_fixed_extensions: List[Tuple[str, Type[BaseProgram]]] = [
    (lang.get("program_about.tab_name"), AboutProgram)
]


def load_extensions():
    if not _extensions:
        _extensions.extend(_fixed_extensions)

        extensions = []

        for _, module_name, _ in pkgutil.iter_modules(
            programs.__path__, f"{programs.__name__}."
        ):
            extensions.append(load_extension(module_name))

        _extensions.extend(
            sorted((ext for ext in extensions if ext is not None), key=lambda x: x[0])
        )


def load_extension(
    module_name: str,
) -> Optional[Tuple[str, Union[Type[BaseProgram], Type[wx.Window]]]]:
    # load module and confirm that all required attributes are defined
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        log.warning(f"Failed to import {module_name}.\n{traceback.format_exc()}")
    else:
        if hasattr(module, "export"):
            export = getattr(module, "export")
            if isinstance(export, dict):
                ui = export.get("ui", None)
                name = export.get("name", None)
                if (
                    isinstance(name, str)
                    and issubclass(ui, BaseProgram)
                    and issubclass(ui, wx.Window)
                ):
                    return name, ui


class WorldPageUI(wx.Notebook, BasePageUI):
    def __init__(
        self, parent: wx.Window, path: str, close_self_callback: Callable[[], None]
    ):
        super().__init__(parent, style=wx.NB_LEFT)
        self._path = path
        self._close_self_callback = close_self_callback
        try:
            self.world = load_level(path)
        except LoaderNoneMatched as e:
            self.Destroy()
            raise e
        self.world_name = self.world.level_wrapper.level_name
        self._extensions: List[BaseProgram] = []
        self._active_extension: int = -1
        self._load_extensions()
        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self._page_change)

    @property
    def path(self) -> str:
        return self._path

    def menu(self, menu: MenuData) -> MenuData:
        menu.setdefault(lang.get("menu_bar.file.menu_name"), {}).setdefault(
            "exit", {}
        ).setdefault(
            lang.get("menu_bar.file.close_world"), lambda evt: self._close_self_callback
        )
        return self._extensions[self.GetSelection()].menu(menu)

    def _load_extensions(self):
        """Load and create instances of each of the extensions"""
        load_extensions()
        select = True
        for extension_name, extension in _extensions:
            try:
                ext = extension(self, self.world, self._close_self_callback)
                self._extensions.append(ext)
                self.AddPage(ext, extension_name, select)
                select = False
            except Exception as e:
                log.exception(
                    f"Failed to load extension {extension_name}\n{e}\n{traceback.format_exc()}"
                )
                continue

    def is_closeable(self) -> bool:
        """Check if all extensions are safe to be closed"""
        return all(e.is_closeable() for e in self._extensions)

    def close(self):
        """Close the world and destroy the UI
        Check is_closeable before running this"""
        for ext in self._extensions:
            ext.close()

        # close the world in a new thread
        thread = threading.Thread(target=self.world.close)
        thread.start()
        # sleep a little
        thread.join(0.1)
        if thread.is_alive():
            # if not closed yet open a dialog to warn the user.
            # We do this on a delay so that it does not flick up for a split second
            dialog = wx.ProgressDialog(
                "Closing World",
                "Please be patient. This may take a little while.",
                maximum=100,
                style=wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME | wx.PD_AUTO_HIDE,
            )
            dialog.Fit()
            dialog.Update(99)
            # wait until the world is closed then close the dialog
            while thread.is_alive():
                wx.GetApp().Yield()
                thread.join(0.1)
            dialog.Destroy()

    def _page_change(self, _):
        """Method to fire when the page is changed"""
        if self.GetSelection() != self._active_extension:
            self._enable_active()

    def _disable_active(self):
        if self._active_extension >= 0:
            try:
                self._extensions[self._active_extension].disable()
            except Exception as e:
                log.critical(traceback.format_exc())
            finally:
                self._active_extension = -1

    def _enable_active(self):
        self._disable_active()
        try:
            self._extensions[self.GetSelection()].enable()
            self.GetGrandParent().create_menu()
        except Exception as e:
            log.critical(traceback.format_exc())
            dialog = TracebackDialog(
                self, "Exception loading sub-program", str(e), traceback.format_exc()
            )
            dialog.ShowModal()
            dialog.Destroy()
            self._extensions.pop(self.GetSelection())
            self._active_extension = -1
            self.DeletePage(self.GetSelection())
        else:
            self._active_extension = self.GetSelection()

    def disable(self):
        """Disable all containers in the world page"""
        self._disable_active()

    def enable(self):
        """Enable the world page"""
        self._enable_active()



from typing import TYPE_CHECKING
import wx
import os

from amulet.api.selection import SelectionGroup
from amulet.api.errors import ChunkLoadError
from amulet.api.data_types import Dimension, OperationReturnType
from amulet.level.formats.schematic import SchematicFormatWrapper

from amulet_map_editor.api.wx.ui.version_select import PlatformSelect
from amulet_map_editor.programs.edit.api.operations import (
    SimpleOperationPanel,
    OperationError,
)

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas

WarningMsg = """The Schematic format is a
legacy format that can only
save data in the numerical
format. Anything that was
added to the game in version
1.13 or after will not be
saved in the schematic file.

We suggest using the Construction
file format instead."""


class ExportSchematic(SimpleOperationPanel):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleOperationPanel.__init__(self, parent, canvas, world, options_path)

        options = self._load_options({})

        self._file_picker = wx.FilePickerCtrl(
            self,
            path=options.get("path", ""),
            wildcard="Schematic file (*.schematic)|*.schematic",
            style=wx.FLP_USE_TEXTCTRL | wx.FLP_SAVE,
        )
        self._sizer.Add(self._file_picker, 0, wx.ALL | wx.CENTER, 5)
        self._platform_define = PlatformSelect(
            self,
            world.translation_manager,
            options.get("platform", None) or world.level_wrapper.platform,
            allow_universal=False,
        )
        self._sizer.Add(self._platform_define, 0, wx.CENTRE, 5)
        self._sizer.Add(
            wx.StaticText(self, label=WarningMsg, style=wx.ALIGN_CENTRE_HORIZONTAL),
            0,
            wx.ALL | wx.CENTER,
            5,
        )
        self._add_run_button("Export")
        self.Layout()

    def disable(self):
        self._save_options(
            {
                "path": self._file_picker.GetPath(),
                "platform": self._platform_define.platform,
            }
        )

    def _operation(
        self, world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
    ) -> OperationReturnType:
        if len(selection.selection_boxes) == 0:
            raise OperationError("No selection was given to export.")
        elif len(selection.selection_boxes) != 1:
            raise OperationError(
                "The schematic format only supports a single selection box."
            )

        path = self._file_picker.GetPath()
        platform = self._platform_define.platform
        if isinstance(path, str):
            wrapper = SchematicFormatWrapper(path)
            if wrapper.exists:
                response = wx.MessageDialog(
                    self,
                    f"A file is already present at {path}. Do you want to continue?",
                    style=wx.YES | wx.NO,
                ).ShowModal()
                if response == wx.ID_CANCEL:
                    return
            wrapper.create_and_open(platform, (1, 12, 2), selection, True)
            wrapper.translation_manager = world.translation_manager
            wrapper_dimension = wrapper.dimensions[0]
            chunk_count = len(list(selection.chunk_locations()))
            yield 0, f"Exporting {os.path.basename(path)}"
            for chunk_index, (cx, cz) in enumerate(selection.chunk_locations()):
                try:
                    chunk = world.get_chunk(cx, cz, dimension)
                    wrapper.commit_chunk(chunk, wrapper_dimension)
                except ChunkLoadError:
                    continue
                yield (chunk_index + 1) / chunk_count
            wrapper.save()
            wrapper.close()
        else:
            raise OperationError(
                "Please specify a save location and platform in the options before running."
            )


export = {
    "name": "Export Schematic (legacy)",  # the name of the plugin
    "operation": ExportSchematic,  # the UI class to display
}



import wx
from typing import Type, Any, Tuple

from amulet.api.data_types import BlockCoordinates
from amulet_map_editor.api.wx.ui.simple import SimplePanel
from amulet_map_editor.api.wx.util.validators import IntValidator
from amulet_map_editor.api import config
from .events import LocationChangeEvent, TransformChangeEvent


class SelectLocationUI(SimplePanel):
    """A UI element that can be dropped into the EditCanvas and let the user pick a location for a structure.
    This UI does not allow for rotation."""

    def __init__(self, parent: wx.Window):
        SimplePanel.__init__(self, parent)
        self._setup_ui()

    def _add_row(self, label: str, wx_object: Type[wx.Object], **kwargs) -> Any:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.add_object(sizer, 0, wx.ALIGN_CENTER_HORIZONTAL)
        name_text = wx.StaticText(self, label=label)
        sizer.Add(name_text, flag=wx.CENTER | wx.ALL | wx.EXPAND, border=5)
        obj = wx_object(self, **kwargs)
        sizer.Add(obj, flag=wx.CENTER | wx.ALL, border=5)
        return obj

    def _setup_ui(self):
        self._x: wx.SpinCtrl = self._add_row(
            "x", wx.SpinCtrl, min=-30000000, max=30000000
        )
        self._y: wx.SpinCtrl = self._add_row(
            "y", wx.SpinCtrl, min=-30000000, max=30000000
        )
        self._z: wx.SpinCtrl = self._add_row(
            "z", wx.SpinCtrl, min=-30000000, max=30000000
        )
        for ui in (self._x, self._y, self._z):
            ui.SetValidator(IntValidator())

        self._x.Bind(wx.EVT_SPINCTRL, self._on_location_change)
        self._y.Bind(wx.EVT_SPINCTRL, self._on_location_change)
        self._z.Bind(wx.EVT_SPINCTRL, self._on_location_change)

        self._copy_air: wx.CheckBox = self._add_row("Copy Air", wx.CheckBox)
        copy_air = config.get("edit_select_location", {}).get("copy_air", False)
        self._copy_air.SetValue(copy_air)
        self._copy_air.Bind(wx.EVT_CHECKBOX, self._save_config)

        self._copy_water: wx.CheckBox = self._add_row("Copy Water", wx.CheckBox)
        copy_water = config.get("edit_select_location", {}).get("copy_water", True)
        self._copy_water.SetValue(copy_water)
        self._copy_water.Bind(wx.EVT_CHECKBOX, self._save_config)

        self._copy_lava: wx.CheckBox = self._add_row("Copy Lava", wx.CheckBox)
        copy_lava = config.get("edit_select_location", {}).get("copy_lava", True)
        self._copy_lava.SetValue(copy_lava)
        self._copy_lava.Bind(wx.EVT_CHECKBOX, self._save_config)

    @property
    def location(self) -> BlockCoordinates:
        return self._x.GetValue(), self._y.GetValue(), self._z.GetValue()

    @location.setter
    def location(self, location: Tuple[int, int, int]):
        x, y, z = location
        self._x.SetValue(x)
        self._y.SetValue(y)
        self._z.SetValue(z)

    @property
    def copy_air(self) -> bool:
        return self._copy_air.GetValue()

    @copy_air.setter
    def copy_air(self, copy_air: bool):
        self._copy_air.SetValue(copy_air)

    @property
    def copy_water(self) -> bool:
        return self._copy_water.GetValue()

    @copy_water.setter
    def copy_water(self, copy_water: bool):
        self._copy_water.SetValue(copy_water)

    @property
    def copy_lava(self) -> bool:
        return self._copy_lava.GetValue()

    @copy_lava.setter
    def copy_lava(self, copy_lava: bool):
        self._copy_lava.SetValue(copy_lava)

    def _save_config(self, evt):
        select_config = config.get("edit_select_location", {})
        select_config["copy_air"] = self.copy_air
        select_config["copy_water"] = self.copy_water
        select_config["copy_lava"] = self.copy_lava
        config.put("edit_select_location", select_config)

    def _on_location_change(self, evt):
        wx.PostEvent(self, LocationChangeEvent(self.location))
        self._on_transform_change()

    def _on_transform_change(self):
        """The object transform has been changed by the user. Sends"""
        wx.PostEvent(self, TransformChangeEvent(self.location, (0, 0, 0), (1, 1, 1)))



import wx
import wx.lib.scrolledpanel
from typing import Tuple, Type

import PyMCTranslate
from amulet.api.data_types import VersionNumberTuple, PlatformType

from amulet_map_editor.api.wx.ui.base_select import EVT_ITEM_CHANGE, BaseSelect
from amulet_map_editor.api.wx.ui.version_select import (
    VersionSelect,
    EVT_VERSION_CHANGE,
)


class BaseDefine(wx.Panel):
    def __init__(
        self,
        parent,
        translation_manager: PyMCTranslate.TranslationManager,
        select_type: Type[BaseSelect],
        orientation=wx.VERTICAL,
        platform: str = None,
        version_number: Tuple[int, int, int] = None,
        namespace: str = None,
        default_name: str = None,
        show_pick: bool = False,
        **kwargs,
    ):
        super().__init__(parent)

        self._translation_manager = translation_manager
        self._orientation = orientation
        self._sizer = wx.BoxSizer(orientation)
        left_sizer = wx.BoxSizer(wx.VERTICAL)
        if orientation == wx.HORIZONTAL:
            self._sizer.Add(left_sizer, 1, wx.EXPAND)
        else:
            self._sizer.Add(left_sizer, 2, wx.EXPAND)

        self._version_picker = VersionSelect(
            self,
            translation_manager,
            platform,
            version_number,
            **kwargs,
        )
        left_sizer.Add(self._version_picker, 0, wx.EXPAND)
        self._version_picker.Bind(EVT_VERSION_CHANGE, self._on_version_change)

        self._picker = select_type(
            self,
            translation_manager,
            self._version_picker.platform,
            self._version_picker.version_number,
            self._version_picker.force_blockstate,
            namespace,
            default_name,
            show_pick,
        )
        left_sizer.Add(self._picker, 1, wx.EXPAND | wx.TOP, 5)
        self._picker.Bind(EVT_ITEM_CHANGE, self._on_picker_change)

        self.SetSizerAndFit(self._sizer)
        self.Layout()

    def _on_picker_change(self, evt):
        raise NotImplementedError("This method should be overridden in child classes.")

    def _on_version_change(self, evt):
        self._picker.version = (
            evt.platform,
            evt.version_number,
            evt.force_blockstate,
        )
        evt.Skip()

    @property
    def platform(self) -> PlatformType:
        return self._version_picker.platform

    @platform.setter
    def platform(self, platform: PlatformType):
        self._version_picker.platform = platform

    @property
    def version_number(self) -> VersionNumberTuple:
        return self._version_picker.version_number

    @version_number.setter
    def version_number(self, version_number: VersionNumberTuple):
        self._version_picker.version_number = version_number

    @property
    def version(self) -> Tuple[PlatformType, VersionNumberTuple, bool]:
        return self._version_picker.version

    @version.setter
    def version(self, version: Tuple[PlatformType, VersionNumberTuple, bool]):
        self._version_picker.version = version

    @property
    def namespace(self) -> str:
        return self._picker.namespace

    @namespace.setter
    def namespace(self, namespace: str):
        self._picker.namespace = namespace



from __future__ import annotations

from functools import partial
from collections.abc import MutableMapping, MutableSequence

import wx

from amulet_map_editor.api.wx.ui import simple

import amulet_nbt as nbt

from amulet_map_editor.api import image

nbt_resources = image.nbt

NBT_FILE = b"\x0A\x00\x0B\x68\x65\x6C\x6C\x6F\x20\x77\x6F\x72\x6C\x64\x08\x00\x04\x6E\x61\x6D\x65\x00\x09\x42\x61\x6E\x61\x6E\x72\x61\x6D\x61\x00"


class NBTRadioButton(simple.SimplePanel):
    def __init__(self, parent, nbt_tag_class, icon):
        super(NBTRadioButton, self).__init__(parent, wx.HORIZONTAL)

        self.nbt_tag_class = nbt_tag_class

        self.radio_button = wx.RadioButton(self, wx.ID_ANY, wx.EmptyString)
        self.add_object(self.radio_button, 0, wx.ALIGN_CENTER | wx.ALL)

        self.tag_bitmap = wx.StaticBitmap(self, wx.ID_ANY, icon)
        self.tag_bitmap.SetToolTip(nbt_tag_class)

        self.add_object(self.tag_bitmap, 0, wx.ALIGN_CENTER | wx.ALL)

    def GetValue(self):
        return self.radio_button.GetValue()

    def SetValue(self, val):
        self.radio_button.SetValue(val)


class NBTEditor(simple.SimplePanel):
    def __init__(self, parent, nbt_data, root_tag_name="", callback=None):
        super(NBTEditor, self).__init__(parent)

        self.nbt_data = nbt_data

        self.image_list = wx.ImageList(16, 16)
        self.image_map = {
            nbt.TAG_Byte: self.image_list.Add(nbt_resources.nbt_tag_byte.bitmap()),
            nbt.TAG_Short: self.image_list.Add(nbt_resources.nbt_tag_short.bitmap()),
            nbt.TAG_Int: self.image_list.Add(nbt_resources.nbt_tag_int.bitmap()),
            nbt.TAG_Long: self.image_list.Add(nbt_resources.nbt_tag_long.bitmap()),
            nbt.TAG_Float: self.image_list.Add(nbt_resources.nbt_tag_float.bitmap()),
            nbt.TAG_Double: self.image_list.Add(nbt_resources.nbt_tag_double.bitmap()),
            nbt.TAG_String: self.image_list.Add(nbt_resources.nbt_tag_string.bitmap()),
            nbt.TAG_Compound: self.image_list.Add(
                nbt_resources.nbt_tag_compound.bitmap()
            ),
            nbt.NBTFile: self.image_list.ImageCount - 1,
            nbt.TAG_List: self.image_list.Add(nbt_resources.nbt_tag_list.bitmap()),
            nbt.TAG_Byte_Array: self.image_list.Add(
                nbt_resources.nbt_tag_array.bitmap()
            ),
            nbt.TAG_Int_Array: self.image_list.ImageCount - 1,
            nbt.TAG_Long_Array: self.image_list.ImageCount - 1,
        }
        self.other = self.image_map[nbt.TAG_String]

        self.tree = self.build_tree(root_tag_name)
        self.add_object(self.tree, 1, wx.ALL | wx.CENTER | wx.EXPAND)

        button_row = simple.SimplePanel(self, wx.HORIZONTAL)
        self.commit_button = wx.Button(button_row, label="Commit")
        self.cancel_button = wx.Button(button_row, label="Cancel")

        button_row.add_object(self.commit_button, space=0)
        button_row.add_object(self.cancel_button, space=0)

        self.commit_button.Bind(wx.EVT_BUTTON, self.commit)
        self.cancel_button.Bind(wx.EVT_BUTTON, self._close)

        self.add_object(button_row, space=0)

        self.callback = callback

    def commit(self, evt):
        self.callback(self.nbt_data)
        evt.Skip()

    def _close(self, evt):
        self.Close(True)
        evt.Skip()

    def build_tree(self, root_tag_name=""):
        def add_tree_node(_tree: wx.TreeCtrl, _parent, _items):
            for key, value in _items.items():
                new_child = None
                if isinstance(value, MutableMapping):
                    new_child = _tree.AppendItem(_parent, key)
                    add_tree_node(_tree, new_child, value)
                elif isinstance(value, MutableSequence):
                    new_child = _tree.AppendItem(_parent, key)

                    for i, item in enumerate(value):
                        child_child = _tree.AppendItem(new_child, f"{item.value}")
                        tree.SetItemData(child_child, (i, item))
                        tree.SetItemImage(
                            child_child,
                            self.image_map.get(item.__class__, self.other),
                            wx.TreeItemIcon_Normal,
                        )
                else:
                    new_child = _tree.AppendItem(_parent, f"{key}: {value.value}")

                tree.SetItemData(new_child, (key, value))
                tree.SetItemImage(
                    new_child, self.image_map.get(value.__class__, self.other)
                )

        tree = wx.TreeCtrl(self, style=wx.TR_DEFAULT_STYLE | wx.EXPAND | wx.ALL)
        tree.AssignImageList(self.image_list)

        root = tree.AddRoot(root_tag_name)
        tree.SetItemData(root, (root_tag_name, self.nbt_data))
        tree.SetItemImage(
            root,
            self.image_map.get(
                self.nbt_data.__class__, self.image_map[nbt.TAG_Compound]
            ),
            wx.TreeItemIcon_Normal,
        )

        add_tree_node(tree, root, self.nbt_data)

        tree.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, self.tree_right_click)

        return tree

    def _generate_menu(self, include_add_tag=False):
        menu = wx.Menu()

        menu_items = [
            wx.MenuItem(menu, text="Edit Tag", id=wx.ID_ANY),
            wx.MenuItem(menu, text="Delete Tag", id=wx.ID_ANY),
        ]

        if include_add_tag:
            menu_items.insert(0, wx.MenuItem(menu, text="Add Tag", id=wx.ID_ANY))

        for menu_item in menu_items:
            menu.Append(menu_item)

        op_map = {
            item.GetId(): item.GetItemLabelText().split()[0].lower()
            for item in menu_items
        }
        menu.Bind(wx.EVT_MENU, lambda evt: self.popup_menu_handler(op_map, evt))

        return menu

    def tree_right_click(self, evt):
        tag_name, tag_obj = self.tree.GetItemData(evt.GetItem())

        menu = self._generate_menu(
            isinstance(tag_obj, (MutableMapping, MutableSequence))
        )
        self.PopupMenu(menu, evt.GetPoint())
        menu.Destroy()
        evt.Skip()

    def popup_menu_handler(self, op_map, evt):
        op_id = evt.GetId()
        op_name = op_map[op_id]

        if op_name == "add":
            self.add_tag()
        elif op_name == "edit":
            self.edit_tag()
        else:
            selected_tag = self.tree.GetFocusedItem()

            if selected_tag == self.tree.GetRootItem():
                wx.MessageBox("Root nodes cannot be deleted currently", "Info", wx.OK)
                return

            name, _ = self.tree.GetItemData(selected_tag)
            parent = self.tree.GetItemParent(selected_tag)

            if parent == self.tree.GetRootItem():
                parent_data = self.nbt_data
            else:
                _, parent_data = self.tree.GetItemData(parent)

            del parent_data[name]

            self.tree.Delete(selected_tag)

    def add_tag(self):
        selected_tag = self.tree.GetFocusedItem()
        name, data = self.tree.GetItemData(selected_tag)

        def save_func(new_name, new_tag_value, new_tag_type, _):
            tag_type = [
                tag_class
                for tag_class in self.image_map
                if tag_class.__name__ == new_tag_type
            ][0]
            self.nbt_data[new_name] = nbt_tag = tag_type(new_tag_value)

            new_child = self.tree.AppendItem(
                selected_tag, f"{new_name}: {new_tag_value}"
            )
            self.tree.SetItemImage(new_child, self.image_map.get(tag_type, self.other))
            self.tree.SetItemData(new_child, (new_name, nbt_tag))

        add_dialog = EditTagDialog(
            self,
            "",
            nbt.TAG_Byte(0),
            [
                tag_type.__name__
                for tag_type in self.image_map.keys()
                if "TAG_" in tag_type.__name__
            ],
            create=True,
            save_callback=save_func,
        )
        add_dialog.Show()

    def edit_tag(self):
        selected_tag = self.tree.GetFocusedItem()
        name, data = self.tree.GetItemData(selected_tag)

        def save_func(new_name, new_tag_value, new_tag_type, old_name):
            tag_type = [
                tag_class
                for tag_class in self.image_map
                if tag_class.__name__ == new_tag_type
            ][0]

            self.nbt_data[new_name] = nbt_tag = tag_type(new_tag_value)
            self.tree.SetItemImage(
                selected_tag, self.image_map.get(tag_type, self.other)
            )
            self.tree.SetItemData(selected_tag, (new_name, nbt_tag))
            self.tree.SetItemText(selected_tag, f"{new_name}: {new_tag_value}")

            if new_name != old_name:
                del self.nbt_data[old_name]

        edit_dialog = EditTagDialog(
            self,
            name,
            data,
            [
                tag_type.__name__
                for tag_type in self.image_map.keys()
                if "TAG_" in tag_type.__name__
            ],
            save_callback=save_func,
        )
        edit_dialog.Show()


class EditTagDialog(wx.Frame):
    GRID_ROWS = 3
    GRID_COLUMNS = 4

    def __init__(
        self, parent, tag_name, tag, tag_types, create=False, save_callback=None
    ):
        super(EditTagDialog, self).__init__(
            parent, title="Edit NBT Tag", size=(500, 280)
        )

        self.save_callback = save_callback
        self.old_name = tag_name
        self.data_type_func = lambda x: x

        main_panel = simple.SimplePanel(self)

        name_panel = simple.SimplePanel(main_panel, sizer_dir=wx.HORIZONTAL)
        value_panel = simple.SimplePanel(main_panel, sizer_dir=wx.HORIZONTAL)
        tag_type_panel = simple.SimplePanel(main_panel)
        button_panel = simple.SimplePanel(main_panel, sizer_dir=wx.HORIZONTAL)

        name_label = wx.StaticText(name_panel, label="Name: ")
        self.name_field = wx.TextCtrl(name_panel)

        if tag_name == "" and not create:
            self.name_field.Disable()
        else:
            self.name_field.SetValue(tag_name)

        name_panel.add_object(name_label, space=0, options=wx.ALL | wx.CENTER)
        name_panel.add_object(self.name_field, space=1, options=wx.ALL | wx.EXPAND)

        value_label = wx.StaticText(value_panel, label="Value: ")
        self.value_field = wx.TextCtrl(value_panel)

        if isinstance(tag, (nbt.TAG_Compound, nbt.TAG_List)):
            self.value_field.Disable()
        else:
            self.value_field.SetValue(str(tag.value))

        value_panel.add_object(value_label, space=0, options=wx.ALL | wx.CENTER)
        value_panel.add_object(self.value_field, space=1, options=wx.ALL | wx.EXPAND)

        tag_type_sizer = wx.GridSizer(self.GRID_ROWS, self.GRID_COLUMNS, 0, 0)

        self.radio_buttons = []

        for tag_type in tag_types:
            rd_btn = NBTRadioButton(
                tag_type_panel,
                tag_type,
                parent.image_list.GetBitmap(parent.image_map[getattr(nbt, tag_type)]),
            )
            self.radio_buttons.append(rd_btn)
            rd_btn.Bind(wx.EVT_RADIOBUTTON, partial(self.handle_radio_button, tag_type))
            tag_type_sizer.Add(rd_btn, 0, wx.ALL, 0)

            if tag_type == tag.__class__.__name__:
                rd_btn.SetValue(True)

        tag_type_panel.SetSizerAndFit(tag_type_sizer)

        self.save_button = wx.Button(button_panel, label="Save")
        self.cancel_button = wx.Button(button_panel, label="Cancel")

        button_panel.add_object(self.save_button, space=0)
        button_panel.add_object(self.cancel_button, space=0)

        main_panel.add_object(name_panel, space=0, options=wx.ALL | wx.EXPAND)
        main_panel.add_object(value_panel, space=0, options=wx.ALL | wx.EXPAND)
        main_panel.add_object(tag_type_panel, space=0)
        main_panel.add_object(button_panel, space=0)

        self.save_button.Bind(wx.EVT_BUTTON, self.save)
        self.cancel_button.Bind(wx.EVT_BUTTON, lambda evt: self.Close())

        self.value_field.Bind(wx.EVT_TEXT, self.value_changed)

        self.SetSize((235, 260))
        self.Layout()

    def value_changed(self, evt):
        tag_value = evt.GetString()
        self.value_field.ChangeValue(str(self.data_type_func(tag_value)))

    def handle_radio_button(self, tag_type, evt):
        for rd_btn in self.radio_buttons:
            rd_btn.SetValue(rd_btn.nbt_tag_class == tag_type)
        self.change_tag_type_func(tag_type)

    def change_tag_type_func(self, tag_type):
        self.data_type_func = lambda x: x
        if tag_type in ("TAG_Int", "TAG_Long", "TAG_Short", "TAG_Short", "TAG_Byte"):
            self.data_type_func = lambda x: int(float(x))
        elif tag_type in ("TAG_Float", "TAG_Double"):
            self.data_type_func = float

        if tag_type not in ("TAG_Byte_Array", "TAG_Int_Array", "TAG_Long_Array"):
            self.value_field.ChangeValue(
                str(self.data_type_func(self.value_field.GetValue()))
            )

    def get_selected_tag_type(self):
        for rd_btn in self.radio_buttons:
            if rd_btn.GetValue():
                return rd_btn.nbt_tag_class
        return None

    def save(self, evt):
        self.save_callback(
            self.name_field.GetValue(),
            self.data_type_func(self.value_field.GetValue()),
            self.get_selected_tag_type(),
            self.old_name,
        )

        self.Close()


if __name__ == "__main__":
    import wx.lib.inspection

    app = wx.App()
    wx.lib.inspection.InspectionTool().Show()
    frame = wx.Frame(None)
    NBTEditor(frame, nbt.load(NBT_FILE), callback=lambda nbt_data: print(nbt_data))
    frame.Show()
    app.MainLoop()



from amulet_map_editor.api.opengl.resource_pack import OpenGLResourcePack


class OpenGLResourcePackManagerStatic:
    """A class to hold an opengl resource pack.
    All classes wanting to access block geometry or textures should subclass this class."""

    def __init__(
        self,
        resource_pack: OpenGLResourcePack,
    ):
        self._resource_pack = resource_pack

    @property
    def resource_pack(self) -> OpenGLResourcePack:
        return self._resource_pack


class OpenGLResourcePackManager(OpenGLResourcePackManagerStatic):
    """A class to hold and enable switching of an opengl resource pack.
    All classes wanting to access block geometry or textures should subclass this class."""

    @property
    def resource_pack(self) -> OpenGLResourcePack:
        return self._resource_pack

    @resource_pack.setter
    def resource_pack(self, resource_pack: OpenGLResourcePack):
        assert isinstance(resource_pack, OpenGLResourcePack)
        self._resource_pack = resource_pack
        self._rebuild()

    def _rebuild(self):
        """Rebuild all geometry."""
        raise NotImplementedError



from typing import TYPE_CHECKING
import wx
from OpenGL.GL import (
    glClear,
    GL_DEPTH_BUFFER_BIT,
)

from .operation_ui import OperationUI
from amulet_map_editor.programs.edit.api.behaviour import StaticSelectionBehaviour
from amulet_map_editor.api.opengl.camera import Projection
from amulet_map_editor.programs.edit.api.events import EVT_DRAW
from amulet_map_editor.programs.edit.api.behaviour import (
    CameraBehaviour,
    PointerBehaviour,
)
from amulet_map_editor.programs.edit.api.events import (
    InputPressEvent,
    EVT_INPUT_PRESS,
)
from amulet_map_editor.programs.edit.api.key_config import (
    ACT_BOX_CLICK,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas
    from amulet.api.level import BaseLevel


class DefaultOperationUI(OperationUI):
    """An extension of the base OperationUI that adds camera, static selection and some other controls."""

    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        super().__init__(parent, canvas, world, options_path)
        self._selection = StaticSelectionBehaviour(self.canvas)
        self._camera_behaviour = CameraBehaviour(self.canvas)
        self._pointer = PointerBehaviour(self.canvas)
        self._show_pointer = False

    def enable(self):
        self._selection.update_selection()
        self.canvas.camera.projection_mode = Projection.PERSPECTIVE

    def bind_events(self):
        self._selection.bind_events()
        self.canvas.Bind(EVT_DRAW, self._on_draw)
        self._camera_behaviour.bind_events()
        self._pointer.bind_events()
        self.canvas.Bind(EVT_INPUT_PRESS, self._on_input_press)

    def _on_draw(self, evt):
        self.canvas.renderer.start_draw()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.canvas.renderer.draw_sky_box()
            glClear(GL_DEPTH_BUFFER_BIT)
        self.canvas.renderer.draw_level()
        self._selection.draw()
        if self._show_pointer:
            self._pointer.draw()
        self.canvas.renderer.end_draw()

    def _on_input_press(self, evt: InputPressEvent):
        if evt.action_id == ACT_BOX_CLICK:
            self._on_box_click()
        evt.Skip()

    def _on_box_click(self):
        pass



import wx

SpecialChrs = {
    wx.WXK_BACK,
    wx.WXK_DELETE,
    wx.WXK_SHIFT,
    wx.WXK_END,
    wx.WXK_HOME,
    wx.WXK_LEFT,
    wx.WXK_RIGHT,
}


class BaseValidator(wx.Validator):
    """Validates data as it is entered into the text controls."""

    def __init__(self):
        super().__init__()
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self):
        return self.__class__()

    def Validate(self, win):
        return True

    def TransferToWindow(self):
        return True

    def TransferFromWindow(self):
        return True

    def OnChar(self, event):
        event.Skip()


class IntValidator(BaseValidator):
    def OnChar(self, event):
        keycode = int(event.GetKeyCode())
        if keycode in SpecialChrs or 48 <= keycode <= 57 or keycode == 45:
            event.Skip()


class FloatValidator(BaseValidator):
    def OnChar(self, event):
        keycode = int(event.GetKeyCode())
        if keycode in SpecialChrs or 45 <= keycode <= 57:
            event.Skip()



from typing import List
from setuptools import setup, find_packages
from wheel.bdist_wheel import bdist_wheel
from Cython.Build import cythonize
import os
import glob
import shutil
import sys
import numpy
import subprocess

try:
    import versioneer
except ImportError:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)))
    import versioneer

# there were issues with other builds carrying over their cache
for d in glob.glob("*.egg-info"):
    shutil.rmtree(d)


def load_requirements(path: str) -> List[str]:
    requirements = []
    with open(path) as f:
        for line in f.readlines():
            line = line.strip()
            if (
                not line
                or line.startswith("#")
                or line.startswith("git+")
                or line.startswith("https:")
            ):
                continue
            elif line.startswith("-r "):
                requirements += load_requirements(line[3:])
            else:
                requirements.append(line)
    return requirements


required_packages = load_requirements("./requirements.txt")
first_party = {
    "amulet-core",
    "amulet-nbt",
    "pymctranslate",
    "minecraft-resource-pack",
}


def freeze_requirements(packages: List[str]) -> List[str]:
    # Pip install the requirements to find the newest compatible versions
    # This makes sure that the source versions are using the same dependencies as the compiled version.
    # This also makes sure that the source version is using the newest version of the dependency.
    if any("~=" in r and r.split("~=", 1)[0].lower() in first_party for r in packages):
        print("pip-install")
        try:
            # make sure pip is up to date
            subprocess.run([sys.executable, "-m", "pip", "install", "--upgrade", "pip"])
            # run pip install
            subprocess.run(
                [sys.executable, "-m", "pip", "install", *packages, "--upgrade"]
            )
            # run pip freeze
            installed = (
                subprocess.check_output(
                    [sys.executable, "-m", "pip", "freeze"], encoding="utf-8"
                )
                .strip()
                .split("\n")
            )
            requirements_map = {r.split("==")[0].lower(): r for r in installed}

            print(installed, requirements_map)
            for index, requirement in enumerate(packages):
                if "~=" in requirement:
                    lib = requirement.split("~=")[0].strip().lower()
                    if lib in first_party and lib in requirements_map:
                        packages[index] = requirements_map[lib]
            print(f"Modified packages to {packages}")
        except Exception as e:
            print("Failed to bake versions:", e)
    return packages


# build cython extensions
if next(glob.iglob("amulet_map_editor/**/*.pyx", recursive=True), None):
    # This throws an error if it does not match any files
    ext = cythonize("amulet_map_editor/**/*.pyx")
else:
    ext = ()


cmdclass = versioneer.get_cmdclass()


# There might be a better way of doing this
# The extra argument needs to be defined in the sdist
# so that it doesn't error. It doesn't actually use it.
class SDist(cmdclass["sdist"]):
    user_options = cmdclass["sdist"].user_options + [
        ("find-libs=", None, ""),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.find_libs = None


class BDistWheel(bdist_wheel):
    user_options = bdist_wheel.user_options + [
        (
            "find-libs=",
            None,
            "Find and fix the newest version of first party libraries. Only used internally.",
        ),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.find_libs = None

    def finalize_options(self):
        if self.find_libs:
            self.distribution.install_requires = freeze_requirements(
                list(self.distribution.install_requires)
            )
        super().finalize_options()


cmdclass["sdist"] = SDist
cmdclass["bdist_wheel"] = BDistWheel


setup(
    install_requires=required_packages,
    packages=find_packages(),
    include_package_data=True,
    cmdclass=cmdclass,
    ext_modules=ext,
    include_dirs=[numpy.get_include()],
)



from typing import Tuple
from amulet_map_editor.api.opengl.mesh.selection import (
    RenderSelection,
    RenderSelectionGroup,
)


class GreenRenderSelection(RenderSelection):
    @property
    def box_tint(self) -> Tuple[float, float, float]:
        return 0.7, 1.0, 0.7


class GreenRenderSelectionGroup(RenderSelectionGroup):
    def _new_render_selection(self):
        return GreenRenderSelection(self._context_identifier, self.resource_pack)



import wx
import traceback
from OpenGL.GL import (
    glClearColor,
    glViewport,
)
import os
from typing import Optional, Generator
import weakref
import sys

from minecraft_model_reader.api.resource_pack.java.download_resources import (
    get_java_vanilla_latest_iter,
    get_java_vanilla_fix,
)
from minecraft_model_reader.api.resource_pack.java import (
    JavaResourcePack,
)
from minecraft_model_reader.api.resource_pack.bedrock.download_resources import (
    get_bedrock_vanilla_latest_iter,
    get_bedrock_vanilla_fix,
)
from minecraft_model_reader.api.resource_pack.bedrock import BedrockResourcePack
from minecraft_model_reader.api.resource_pack import (
    load_resource_pack,
    load_resource_pack_manager,
)
from amulet.api.player import LOCAL_PLAYER

from amulet.api.data_types import OperationYieldType, Dimension

from amulet_map_editor import experimental_bedrock_resources
from amulet_map_editor.api.opengl.canvas import EventCanvas
from amulet_map_editor.api.opengl.resource_pack.resource_pack import OpenGLResourcePack
from amulet_map_editor.programs.edit.api.selection import (
    SelectionManager,
    SelectionHistoryManager,
)
from amulet_map_editor import log, lang
import amulet_map_editor.programs.edit as amulet_edit

from amulet_map_editor.api.opengl.camera import ControllableCamera
from amulet_map_editor.api.wx.util.button_input import ButtonInput
from amulet_map_editor.api.wx.util.mouse_movement import MouseMovement
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from ..renderer import Renderer

from amulet.api.level import BaseLevel


class BaseEditCanvas(EventCanvas):
    """Adds base logic for drawing everything related to the edit program to the canvas.
    All the user interaction code is implemented in ControllableEditCanvas to make them easier to read."""

    background_colour = (0.61, 0.70, 0.85)

    def __init__(self, parent: wx.Window, world: BaseLevel):
        super().__init__(parent)
        glClearColor(*self.background_colour, 1.0)
        self.Hide()

        self._world = weakref.ref(world)

        self._selection: Optional[SelectionHistoryManager] = SelectionHistoryManager(
            SelectionManager(self)
        )
        self.world.history_manager.register(self._selection, False)

        self._camera: ControllableCamera = ControllableCamera(self)

        self._camera.move_speed = 2.0
        self._camera.rotate_speed = 2.0

        self._buttons: ButtonInput = ButtonInput(self)
        self._mouse: MouseMovement = MouseMovement(self)
        self._mouse.set_middle()

        # load the resource packs
        try:
            os.makedirs("resource_packs", exist_ok=True)
            if not os.path.isfile("resource_packs/readme.txt"):
                with open("resource_packs/readme.txt", "w") as f:
                    f.write("Put the Java resource pack you want loaded in here.")
        except PermissionError as e:
            raise PermissionError(
                "Amulet is not able to write to the install directory. Try moving Amulet to somewhere else on your computer."
            ) from e

        self._renderer: Optional[Renderer] = None

    def __setattr__(self, name, value):
        if not self.__dict__.get("_init", False) or hasattr(self, name):
            super().__setattr__(name, value)
        else:
            print(name, value)
            traceback.print_stack()

    def setup(self) -> Generator[OperationYieldType, None, None]:
        """Set up objects that take a while to set up.
        If you want to extend setup subclass the _setup method not this method."""
        yield from self._setup()
        self._finalise()

    def _setup(self) -> Generator[OperationYieldType, None, None]:
        """Set up objects that take a while to set up."""
        packs = []
        user_packs = [
            load_resource_pack(os.path.join("resource_packs", rp))
            for rp in os.listdir("resource_packs")
            if os.path.isdir(os.path.join("resource_packs", rp))
        ]
        if (
            self.world.level_wrapper.platform == "bedrock"
            and experimental_bedrock_resources
        ):
            packs.append(
                load_resource_pack(
                    os.path.join(
                        os.path.dirname(amulet_edit.__file__),
                        "amulet_resource_pack",
                        "bedrock",
                    )
                )
            )
            yield 0.1, lang.get(
                "program_3d_edit.canvas.downloading_bedrock_vanilla_resource_pack"
            )
            gen = get_bedrock_vanilla_latest_iter()
            try:
                while True:
                    yield next(gen) * 0.4 + 0.1
            except StopIteration as e:
                packs.append(e.value)
            yield 0.5, lang.get("program_3d_edit.canvas.loading_resource_packs")

            packs += [
                pack for pack in user_packs if isinstance(pack, BedrockResourcePack)
            ]
            packs.append(get_bedrock_vanilla_fix())

            translator = self.world.translation_manager.get_version(
                "bedrock", (999, 0, 0)
            )
        else:
            packs.append(
                load_resource_pack(
                    os.path.join(
                        os.path.dirname(amulet_edit.__file__),
                        "amulet_resource_pack",
                        "java",
                    )
                )
            )
            yield 0.1, lang.get(
                "program_3d_edit.canvas.downloading_java_vanilla_resource_pack"
            )
            gen = get_java_vanilla_latest_iter()
            try:
                while True:
                    yield next(gen) * 0.4 + 0.1
            except StopIteration as e:
                packs.append(e.value)
            except Exception as e:
                if sys.platform == "darwin" and "CERTIFICATE_VERIFY_FAILED" in str(e):
                    msg = lang.get(
                        "program_3d_edit.canvas.java_rp_failed_mac_certificates"
                    )
                else:
                    msg = lang.get("program_3d_edit.canvas.java_rp_failed_default")
                log.error(
                    msg,
                    exc_info=True,
                )
                dialog = TracebackDialog(
                    self,
                    lang.get("program_3d_edit.canvas.java_rp_failed"),
                    f"{msg}\n{e}",
                    traceback.format_exc(),
                )
                dialog.ShowModal()
                dialog.Destroy()

            yield 0.5, lang.get("program_3d_edit.canvas.loading_resource_packs")
            packs += [pack for pack in user_packs if isinstance(pack, JavaResourcePack)]
            packs.append(get_java_vanilla_fix())

            translator = self.world.translation_manager.get_version("java", (999, 0, 0))

        resource_pack = load_resource_pack_manager(packs, load=False)
        for i in resource_pack.reload():
            yield i / 4 + 0.5

        opengl_resource_pack = OpenGLResourcePack(resource_pack, translator)

        yield 0.75, lang.get("program_3d_edit.canvas.creating_texture_atlas")
        for i in opengl_resource_pack.setup():
            yield i / 4 + 0.75

        yield 1.0, lang.get("program_3d_edit.canvas.setting_up_renderer")

        self._renderer: Optional[Renderer] = Renderer(
            self,
            self.world,
            self.context_identifier,
            opengl_resource_pack,
        )

    def _finalise(self):
        """Any logic that needs to be run after everything has been set up."""
        self._init = True
        try:
            player = self.world.get_player(LOCAL_PLAYER)
            self._camera.location_rotation = player.location, player.rotation
            self.dimension = player.dimension
        except:
            self._camera.location_rotation = (0.0, 100.0, 0.0), (45.0, 45.0)

    def bind_events(self):
        """Set up all events required to run.
        Note this will also bind subclass events."""
        self.Bind(wx.EVT_SIZE, self._on_size)
        self.selection.bind_events()
        self.buttons.bind_events()
        self.mouse.bind_events()
        self.renderer.bind_events()

    def enable(self):
        """Enable the canvas and start it working."""
        self.SetCurrent(self._context)
        self.renderer.enable()
        self.buttons.enable()

    def disable(self):
        """Disable the canvas and unload all geometry."""
        self.renderer.disable()
        self.buttons.disable()

    def is_closeable(self):
        """Check that the canvas and contained data is safe to be closed."""
        return self.renderer.is_closeable()

    def close(self):
        """Destroy all contained data so that the window can be safely destroyed."""
        self.renderer.close()
        super().close()

    @property
    def world(self) -> BaseLevel:
        """The amulet-core `BaseLevel` subclass representing the "world" that is open."""
        return self._world()

    @property
    def renderer(self) -> Renderer:
        """The `Renderer` class that handles the drawable objects and draws them."""
        return self._renderer

    @property
    def dimension(self) -> Dimension:
        """The currently loaded dimension in the renderer."""
        return self.renderer.dimension

    @dimension.setter
    def dimension(self, dimension: Dimension):
        """Set the currently loaded dimension in the renderer."""
        self.renderer.dimension = dimension

    @property
    def camera(self) -> ControllableCamera:
        """A class holding the state of the camera with methods to access and modify the state."""
        return self._camera

    @property
    def buttons(self) -> ButtonInput:
        """A class that manages converting button presses into actions based on set keybinds."""
        return self._buttons

    @property
    def mouse(self) -> MouseMovement:
        """A class that manages mouse movement."""
        return self._mouse

    @property
    def selection(self) -> SelectionManager:
        """A simple class for storing the selection state."""
        return self._selection.value

    def _on_size(self, evt):
        wx.CallAfter(self._set_size)
        evt.Skip()

    def _set_size(self):
        size = self.GetClientSize() * self.GetContentScaleFactor()
        width = size.width
        height = size.height
        self.SetCurrent(self._context)
        glViewport(0, 0, width, height)
        if height > 0:
            self.camera.aspect_ratio = width / height
        else:
            self.camera.aspect_ratio = 1



from typing import TYPE_CHECKING
import wx
from wx.adv import RichToolTip
from amulet_map_editor import log
from amulet.api.chunk.biomes import BiomesShape

from .base_behaviour import BaseBehaviour
from .pointer_behaviour import PointerBehaviour
from ..events import (
    InputPressEvent,
    EVT_INPUT_PRESS,
)
from ..key_config import (
    ACT_INSPECT_BLOCK,
)

if TYPE_CHECKING:
    from ..canvas import EditCanvas


class InspectBlockBehaviour(BaseBehaviour):
    """Adds a popup with block information."""

    def __init__(self, canvas: "EditCanvas", pointer_behaviour: PointerBehaviour):
        super().__init__(canvas)
        self._pointer_behaviour = pointer_behaviour

    def bind_events(self):
        """Set up all events required to run."""
        self.canvas.Bind(EVT_INPUT_PRESS, self._on_input_press)

    def _on_input_press(self, evt: InputPressEvent):
        """Logic to run each time the input press event is run."""
        if evt.action_id == ACT_INSPECT_BLOCK:
            self._inspect_block()
        evt.Skip()

    def _inspect_block(self):
        def truncate(s: str, max_line_length: int = None) -> str:
            if isinstance(max_line_length, int):
                max_line_length = max(-1, max_line_length)
                s = "\n".join(
                    [
                        line[: max_line_length - 3] + "..."
                        if len(line) > max_line_length
                        else line
                        for line in s.split("\n")
                    ]
                )
            return s

        full_msg = self._get_block_info_message()
        msg = truncate(full_msg, 150)
        tooltip = RichToolTip("Inspect Block", msg)
        x, y = self.canvas.mouse.xy
        tooltip.ShowFor(self.canvas, wx.Rect(x, y, 1, 1))

    def _get_block_info_message(self) -> str:
        x, y, z = self._pointer_behaviour.pointer_base
        try:
            block = self.canvas.world.get_block(x, y, z, self.canvas.dimension)
            chunk = self.canvas.world.get_chunk(x >> 4, z >> 4, self.canvas.dimension)
            block_entity = chunk.block_entities.get((x, y, z), None)
            platform = self.canvas.world.level_wrapper.platform
            version = self.canvas.world.level_wrapper.version
            translator = self.canvas.world.translation_manager.get_version(
                platform,
                version,
            )
            (version_block, version_block_entity, _,) = translator.block.from_universal(
                block, block_entity, block_location=(x, y, z)
            )
            if isinstance(version, tuple):
                version_str = ".".join(str(v) for v in version[:4])
            else:
                version_str = str(version)
            block_data_text = f"x: {x}, y: {y}, z: {z}\n\n{platform.capitalize()} {version_str}\n{version_block}"
            if version_block_entity:
                version_block_entity_str = str(version_block_entity)
                block_data_text = f"{block_data_text}\n{version_block_entity_str}"

            block_data_text = f"{block_data_text}\n\nUniversal\n{block}"
            if block_entity:
                block_entity_str = str(block_entity)
                block_data_text = f"{block_data_text}\n{block_entity_str}"

            if chunk.biomes.dimension == BiomesShape.Shape2D:
                biome = chunk.biomes[x % 16, z % 16]
                try:
                    block_data_text = f"{block_data_text}\n\nBiome: {self.canvas.world.biome_palette[biome]}"
                except Exception as e:
                    log.error(e)
            elif chunk.biomes.dimension == BiomesShape.Shape3D:
                biome = chunk.biomes[(x % 16) // 4, y // 4, (z % 16) // 4]
                try:
                    block_data_text = f"{block_data_text}\n\nBiome: {self.canvas.world.biome_palette[biome]}"
                except Exception as e:
                    log.error(e)

        except Exception as e:
            log.error(e)
            return str(e)
        else:
            return block_data_text
